import logging
import multiprocessing
import os
from concurrent.futures import ThreadPoolExecutor

import grpc
from google.protobuf import empty_pb2 as _empty_pb2
from pynumaflow.info.server import get_sdk_version, write as info_server_write
from pynumaflow.info.types import ServerInfo, Protocol, Language, SERVER_INFO_FILE_PATH

from pynumaflow import setup_logging
from pynumaflow._constants import (
    MAX_MESSAGE_SIZE,
    MAP_SOCK_PATH,
)
from pynumaflow.mapper import Datum
from pynumaflow.mapper._dtypes import MapCallable
from pynumaflow.mapper.proto import map_pb2
from pynumaflow.mapper.proto import map_pb2_grpc
from pynumaflow.types import NumaflowServicerContext

_LOGGER = setup_logging(__name__)
if os.getenv("PYTHONDEBUG"):
    _LOGGER.setLevel(logging.DEBUG)


_PROCESS_COUNT = multiprocessing.cpu_count()
MAX_THREADS = int(os.getenv("MAX_THREADS", 0)) or (_PROCESS_COUNT * 4)


class Mapper(map_pb2_grpc.MapServicer):
    """
    Provides an interface to write a Mapper
    which will be exposed over a Synchronous gRPC server.

    Args:
        handler: Function callable following the type signature of MapCallable
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
                     defaults to number of processors x4

    Example invocation:
    >>> from typing import Iterator
    >>> from pynumaflow.mapper import Messages, Message\
    ...     Datum, Mapper
    ...
    >>> def map_handler(key: [str], datum: Datum) -> Messages:
    ...   val = datum.value
    ...   _ = datum.event_time
    ...   _ = datum.watermark
    ...   messages = Messages(Message(val, keys=keys))
    ...   return messages
    ...
    >>> grpc_server = Mapper(handler=map_handler)
    >>> grpc_server.start()
    """

    def __init__(
        self,
        handler: MapCallable,
        sock_path=MAP_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=MAX_THREADS,
    ):
        self.__map_handler: MapCallable = handler
        self.sock_path = f"unix://{sock_path}"
        self._max_message_size = max_message_size
        self._max_threads = max_threads
        self.cleanup_coroutines = []

        self._server_options = [
            ("grpc.max_send_message_length", self._max_message_size),
            ("grpc.max_receive_message_length", self._max_message_size),
        ]

    def MapFn(
        self, request: map_pb2.MapRequest, context: NumaflowServicerContext
    ) -> map_pb2.MapResponse:
        """
        Applies a function to each datum element.
        The pascal case function name comes from the proto map_pb2_grpc.py file.
        """
        # proto repeated field(keys) is of type google._upb._message.RepeatedScalarContainer
        # we need to explicitly convert it to list
        try:
            msgs = self.__map_handler(
                list(request.keys),
                Datum(
                    keys=list(request.keys),
                    value=request.value,
                    event_time=request.event_time.ToDatetime(),
                    watermark=request.watermark.ToDatetime(),
                ),
            )
        except Exception as err:
            _LOGGER.critical("UDFError, re-raising the error", exc_info=True)
            context.set_code(grpc.StatusCode.UNKNOWN)
            context.set_details(str(err))
            return map_pb2.MapResponse(results=[])

        datums = []

        for msg in msgs:
            datums.append(map_pb2.MapResponse.Result(keys=msg.keys, value=msg.value, tags=msg.tags))

        return map_pb2.MapResponse(results=datums)

    def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> map_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the proto map_pb2_grpc.py file.
        """
        return map_pb2.ReadyResponse(ready=True)

    def start(self) -> None:
        """
        Starts the gRPC server on the given UNIX socket with given max threads.
        """
        server = grpc.server(
            ThreadPoolExecutor(max_workers=self._max_threads), options=self._server_options
        )
        map_pb2_grpc.add_MapServicer_to_server(self, server)
        server.add_insecure_port(self.sock_path)
        server.start()
        serv_info = ServerInfo(
            protocol=Protocol.UDS,
            language=Language.PYTHON,
            version=get_sdk_version(),
        )
        info_server_write(server_info=serv_info, info_file=SERVER_INFO_FILE_PATH)
        _LOGGER.info(
            "GRPC Server listening on: %s with max threads: %s", self.sock_path, self._max_threads
        )
        server.wait_for_termination()



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from . import mapstream_pb2 as mapstream__pb2


class MapStreamStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.MapStreamFn = channel.unary_stream(
            "/mapstream.v1.MapStream/MapStreamFn",
            request_serializer=mapstream__pb2.MapStreamRequest.SerializeToString,
            response_deserializer=mapstream__pb2.MapStreamResponse.FromString,
        )
        self.IsReady = channel.unary_unary(
            "/mapstream.v1.MapStream/IsReady",
            request_serializer=google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            response_deserializer=mapstream__pb2.ReadyResponse.FromString,
        )


class MapStreamServicer(object):
    """Missing associated documentation comment in .proto file."""

    def MapStreamFn(self, request, context):
        """MapStreamFn applies a function to each request element and returns a stream."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def IsReady(self, request, context):
        """IsReady is the heartbeat endpoint for gRPC."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_MapStreamServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "MapStreamFn": grpc.unary_stream_rpc_method_handler(
            servicer.MapStreamFn,
            request_deserializer=mapstream__pb2.MapStreamRequest.FromString,
            response_serializer=mapstream__pb2.MapStreamResponse.SerializeToString,
        ),
        "IsReady": grpc.unary_unary_rpc_method_handler(
            servicer.IsReady,
            request_deserializer=google_dot_protobuf_dot_empty__pb2.Empty.FromString,
            response_serializer=mapstream__pb2.ReadyResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "mapstream.v1.MapStream", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class MapStream(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def MapStreamFn(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_stream(
            request,
            target,
            "/mapstream.v1.MapStream/MapStreamFn",
            mapstream__pb2.MapStreamRequest.SerializeToString,
            mapstream__pb2.MapStreamResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def IsReady(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/mapstream.v1.MapStream/IsReady",
            google_dot_protobuf_dot_empty__pb2.Empty.SerializeToString,
            mapstream__pb2.ReadyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



import contextlib
import logging
import multiprocessing
import os
import socket
from concurrent import futures

import grpc
from google.protobuf import empty_pb2 as _empty_pb2

from pynumaflow import setup_logging
from pynumaflow._constants import (
    MAX_MESSAGE_SIZE,
)
from pynumaflow._constants import MULTIPROC_MAP_SOCK_PORT, MULTIPROC_MAP_SOCK_ADDR
from pynumaflow.exceptions import SocketError
from pynumaflow.mapper import Datum
from pynumaflow.mapper._dtypes import MapCallable
from pynumaflow.mapper.proto import map_pb2
from pynumaflow.mapper.proto import map_pb2_grpc
from pynumaflow.types import NumaflowServicerContext
from pynumaflow.info.server import (
    get_sdk_version,
    write as info_server_write,
    get_metadata_env,
)
from pynumaflow.info.types import (
    ServerInfo,
    Protocol,
    Language,
    SERVER_INFO_FILE_PATH,
    METADATA_ENVS,
)

_LOGGER = setup_logging(__name__)
if os.getenv("PYTHONDEBUG"):
    _LOGGER.setLevel(logging.DEBUG)


class MultiProcMapper(map_pb2_grpc.MapServicer):
    """
    Provides an interface to write a Multi Proc Mapper
    which will be exposed over gRPC.

    Args:
        handler: Function callable following the type signature of MapCallable
        sock_path: Path to the TCP port to bind to
        max_message_size: The max message size in bytes the server can receive and send

    Example invocation:
    >>> from typing import Iterator
    >>> from pynumaflow.mapper import Messages, Message \
    ...     Datum, MultiProcMapper
    ...
    >>> def map_handler(keys: list[str], datum: Datum) -> Messages:
    ...   val = datum.value
    ...   _ = datum.event_time
    ...   _ = datum.watermark
    ...   messages = Messages(Message(val, keys=keys))
    ...   return messages
    ...
    >>> grpc_server = MultiProcMapper(handler=map_handler)
    >>> grpc_server.start()
    """

    __slots__ = (
        "__map_handler",
        "_max_message_size",
        "_server_options",
        "_sock_path",
        "_process_count",
        "_threads_per_proc",
    )

    def __init__(
        self,
        handler: MapCallable,
        sock_path=MULTIPROC_MAP_SOCK_PORT,
        max_message_size=MAX_MESSAGE_SIZE,
    ):
        self.__map_handler: MapCallable = handler
        self._max_message_size = max_message_size

        self._server_options = [
            ("grpc.max_send_message_length", self._max_message_size),
            ("grpc.max_receive_message_length", self._max_message_size),
            ("grpc.so_reuseport", 1),
            ("grpc.so_reuseaddr", 1),
        ]
        self._sock_path = sock_path
        self._process_count = int(os.getenv("NUM_CPU_MULTIPROC") or os.cpu_count())
        self._threads_per_proc = int(os.getenv("MAX_THREADS", "4"))

    def MapFn(
        self, request: map_pb2.MapRequest, context: NumaflowServicerContext
    ) -> map_pb2.MapResponse:
        """
        Applies a function to each datum element.
        The pascal case function name comes from the proto map_pb2_grpc.py file.
        """
        # proto repeated field(keys) is of type google._upb._message.RepeatedScalarContainer
        # we need to explicitly convert it to list
        try:
            msgs = self.__map_handler(
                list(request.keys),
                Datum(
                    keys=list(request.keys),
                    value=request.value,
                    event_time=request.event_time.ToDatetime(),
                    watermark=request.watermark.ToDatetime(),
                ),
            )
        except Exception as err:
            _LOGGER.critical("UDFError, re-raising the error", exc_info=True)
            context.set_code(grpc.StatusCode.UNKNOWN)
            context.set_details(str(err))
            return map_pb2.MapResponse(results=[])

        datums = []

        for msg in msgs:
            datums.append(map_pb2.MapResponse.Result(keys=msg.keys, value=msg.value, tags=msg.tags))

        return map_pb2.MapResponse(results=datums)

    def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> map_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the proto map_pb2_grpc.py file.
        """
        return map_pb2.ReadyResponse(ready=True)

    def _run_server(self, bind_address: str) -> None:
        """Start a server in a subprocess."""
        _LOGGER.info(
            "Starting new server with num_procs: %s, num_threads/proc: %s",
            self._process_count,
            self._threads_per_proc,
        )
        server = grpc.server(
            futures.ThreadPoolExecutor(
                max_workers=self._threads_per_proc,
            ),
            options=self._server_options,
        )
        map_pb2_grpc.add_MapServicer_to_server(self, server)
        server.add_insecure_port(bind_address)
        server.start()
        serv_info = ServerInfo(
            protocol=Protocol.TCP,
            language=Language.PYTHON,
            version=get_sdk_version(),
            metadata=get_metadata_env(envs=METADATA_ENVS),
        )
        # Overwrite the CPU_LIMIT metadata using user input
        serv_info.metadata["CPU_LIMIT"] = str(self._process_count)
        info_server_write(server_info=serv_info, info_file=SERVER_INFO_FILE_PATH)

        _LOGGER.info("GRPC Multi-Processor Server listening on: %s %d", bind_address, os.getpid())
        server.wait_for_termination()

    @contextlib.contextmanager
    def _reserve_port(self) -> int:
        """Find and reserve a port for all subprocesses to use."""
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if sock.getsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR) == 0:
            raise SocketError("Failed to set SO_REUSEADDR.")
        if sock.getsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT) == 0:
            raise SocketError("Failed to set SO_REUSEPORT.")
        sock.bind(("", self._sock_path))
        try:
            yield sock.getsockname()[1]
        finally:
            sock.close()

    def start(self) -> None:
        """Start N grpc servers in different processes where N = CPU Count"""
        with self._reserve_port() as port:
            bind_address = f"{MULTIPROC_MAP_SOCK_ADDR}:{port}"
            workers = []
            for _ in range(self._process_count):
                # NOTE: It is imperative that the worker subprocesses be forked before
                # any gRPC servers start up. See
                # https://github.com/grpc/grpc/issues/16001 for more details.
                worker = multiprocessing.Process(target=self._run_server, args=(bind_address,))
                worker.start()
                workers.append(worker)
            for worker in workers:
                worker.join()



from pynumaflow.sinker._dtypes import Response, Responses, Datum
from pynumaflow.sinker.async_sink import AsyncSinker
from pynumaflow.sinker.server import Sinker

__all__ = ["Response", "Responses", "Datum", "Sinker", "AsyncSinker"]



from dataclasses import dataclass, field
from enum import Enum

# Constants for using in the info-server
# Need to keep consistent with all SDKs and client
SERVER_INFO_FILE_PATH = "/var/run/numaflow/server-info"
EOF = "U+005C__END__"

# Env variables to be passed in the info server metadata.
# These need to be accessed in the client using the same key.
# Format - (key, env_var)
METADATA_ENVS = [("CPU_LIMIT", "NUMAFLOW_CPU_LIMIT")]


class Protocol(str, Enum):
    """
    Enumerate grpc server connection protocol.
    """

    UDS = "uds"
    TCP = "tcp"


class Language(str, Enum):
    """
    Enumerate Numaflow SDK language.
    """

    GO = "go"
    PYTHON = "python"
    JAVA = "java"


@dataclass
class ServerInfo:
    """
    ServerInfo is used for the gRPC server to provide the information such as protocol,
    sdk version, language, metadata to the client.
    Args:
        protocol: Protocol to use (UDS or TCP)
        language: Language used by the server(Python, Golang, Java)
        version: Numaflow sdk version used by the server
        metadata: Any additional information to be provided (env vars)
    """

    protocol: Protocol
    language: Language
    version: str
    metadata: dict = field(default_factory=dict)



import asyncio

from pynumaflow._constants import STREAM_EOF


class NonBlockingIterator:
    """An Async Interator backed by a queue"""

    __slots__ = "_queue"

    def __init__(self):
        self._queue = asyncio.Queue()

    async def read_iterator(self):
        item = await self._queue.get()
        while True:
            if item == STREAM_EOF:
                break
            yield item
            item = await self._queue.get()

    async def put(self, item):
        await self._queue.put(item)



import logging
import multiprocessing
import os
from concurrent.futures import ThreadPoolExecutor
from typing import Callable

import grpc
from google.protobuf import empty_pb2 as _empty_pb2

from pynumaflow import setup_logging
from pynumaflow._constants import (
    MAX_MESSAGE_SIZE,
    SIDE_INPUT_SOCK_PATH,
)
from pynumaflow.sideinput import Response
from pynumaflow.sideinput.proto import sideinput_pb2, sideinput_pb2_grpc
from pynumaflow.types import NumaflowServicerContext

_LOGGER = setup_logging(__name__)
if os.getenv("PYTHONDEBUG"):
    _LOGGER.setLevel(logging.DEBUG)

RetrieverCallable = Callable[[], Response]
_PROCESS_COUNT = multiprocessing.cpu_count()
MAX_THREADS = int(os.getenv("MAX_THREADS", 0)) or (_PROCESS_COUNT * 4)


class SideInput(sideinput_pb2_grpc.SideInputServicer):
    """
    Provides an interface to write a User Defined Side Input (UDSideInput)
    which will be exposed over gRPC.

    Args:
        handler: Function callable following the type signature of RetrieverCallable
        sock_path: Path to the UNIX Domain Socket
        max_message_size: The max message size in bytes the server can receive and send
        max_threads: The max number of threads to be spawned;
                     defaults to number of processors x 4

    Example invocation:
    >>> from typing import List
    >>> from pynumaflow.sideinput import Response, SideInput
    >>> def my_handler() -> Response:
    ...   response = Response.broadcast_message(b"hello")
    ...   return response
    >>> grpc_server = SideInput(my_handler)
    >>> grpc_server.start()
    """

    SIDE_INPUT_DIR_PATH = "/var/numaflow/side-inputs"

    def __init__(
        self,
        handler: RetrieverCallable,
        sock_path=SIDE_INPUT_SOCK_PATH,
        max_message_size=MAX_MESSAGE_SIZE,
        max_threads=MAX_THREADS,
    ):
        self.__retrieve_handler: RetrieverCallable = handler
        self.sock_path = f"unix://{sock_path}"
        self._max_message_size = max_message_size
        self._max_threads = max_threads
        self.cleanup_coroutines = []

        self._server_options = [
            ("grpc.max_send_message_length", self._max_message_size),
            ("grpc.max_receive_message_length", self._max_message_size),
        ]

    def RetrieveSideInput(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> sideinput_pb2.SideInputResponse:
        """
        Applies a sideinput function for a retrieval request.
        The pascal case function name comes from the proto sideinput_pb2_grpc.py file.
        """
        # if there is an exception, we will mark all the responses as a failure
        try:
            rspn = self.__retrieve_handler()
        except Exception as err:
            err_msg = "RetrieveSideInputErr: %r" % err
            _LOGGER.critical(err_msg, exc_info=True)
            context.set_code(grpc.StatusCode.UNKNOWN)
            context.set_details(str(err))
            return sideinput_pb2.SideInputResponse(value=None, no_broadcast=True)

        return sideinput_pb2.SideInputResponse(value=rspn.value, no_broadcast=rspn.no_broadcast)

    def IsReady(
        self, request: _empty_pb2.Empty, context: NumaflowServicerContext
    ) -> sideinput_pb2.ReadyResponse:
        """
        IsReady is the heartbeat endpoint for gRPC.
        The pascal case function name comes from the proto sideinput_pb2_grpc.py file.
        """
        return sideinput_pb2.ReadyResponse(ready=True)

    def start(self) -> None:
        """
        Starts the gRPC server on the given UNIX socket with given max threads.
        """
        server = grpc.server(
            ThreadPoolExecutor(max_workers=self._max_threads), options=self._server_options
        )
        sideinput_pb2_grpc.add_SideInputServicer_to_server(
            SideInput(self.__retrieve_handler), server
        )
        server.add_insecure_port(self.sock_path)
        server.start()
        _LOGGER.info(
            "Side Input gRPC Server listening on: %s with max threads: %s",
            self.sock_path,
            self._max_threads,
        )
        server.wait_for_termination()



from pynumaflow.sourcetransformer._dtypes import (
    Message,
    Messages,
    Datum,
    DROP,
)
from pynumaflow.sourcetransformer.multiproc_server import MultiProcSourceTransformer
from pynumaflow.sourcetransformer.server import SourceTransformer

__all__ = [
    "Message",
    "Messages",
    "Datum",
    "DROP",
    "SourceTransformer",
    "MultiProcSourceTransformer",
]



import os
from importlib.metadata import version
from typing import Any

from pynumaflow import setup_logging
from pynumaflow.info.types import ServerInfo, EOF
import json
import logging

_LOGGER = setup_logging(__name__)
if os.getenv("PYTHONDEBUG"):
    _LOGGER.setLevel(logging.DEBUG)


def get_sdk_version() -> str:
    """
    Return the pynumaflow SDK version
    """
    try:
        return version("pynumaflow")
    except Exception as e:
        # Adding this to handle the case for local test/CI where pynumaflow
        # will not be installed as a package
        _LOGGER.error("Could not read SDK version %r", e, exc_info=True)
        return ""


def write(server_info: ServerInfo, info_file: str):
    """
    Write the ServerInfo to a file , shared with the client (numa container).

    args:
        serv: The ServerInfo object to be shared
        info_file: the shared file path
    """
    try:
        data = server_info.__dict__
        with open(info_file, "w+") as f:
            json.dump(data, f, ensure_ascii=False)
            f.write(EOF)
    except Exception as err:
        _LOGGER.critical("Could not write data to Info-Server %r", err, exc_info=True)
        raise err


def get_metadata_env(envs: list[tuple[str, str]]) -> dict[str, Any]:
    """
    Extract the environment var value from the provided list,
    and assign them to the given key in the metadata

    args:
        envs: List of tuples (key, env_var)
    """
    meta = {}
    for key, val in envs:
        res = os.getenv(val, None)
        if res:
            meta[key] = res
    return meta



from dataclasses import dataclass
from datetime import datetime
from typing import TypeVar, Optional, Callable
from collections.abc import Sequence, Iterator
from warnings import warn

R = TypeVar("R", bound="Response")
Rs = TypeVar("Rs", bound="Responses")


@dataclass
class Response:
    """
    Basic datatype for UDSink response.

    Args:
        id: the id of the event.
        success: boolean indicating whether the event was successfully processed.
        err: error message if the event was not successfully processed.
    """

    id: str
    success: bool
    err: Optional[str]

    __slots__ = ("id", "success", "err")

    @classmethod
    def as_success(cls: type[R], id_: str) -> R:
        return Response(id=id_, success=True, err=None)

    @classmethod
    def as_failure(cls: type[R], id_: str, err_msg: str) -> R:
        return Response(id=id_, success=False, err=err_msg)


class Responses(Sequence[R]):
    """
    Container to hold a list of Response instances.

    Args:
        responses: list of Response instances.
    """

    __slots__ = ("_responses",)

    def __init__(self, *responses: R):
        self._responses = list(responses) or []

    def __str__(self) -> str:
        return str(self._responses)

    def __repr__(self) -> str:
        return str(self)

    def __len__(self) -> int:
        return len(self._responses)

    def __iter__(self) -> Iterator[R]:
        return iter(self._responses)

    def __getitem__(self, index: int) -> R:
        if isinstance(index, slice):
            raise TypeError("Slicing is not supported for Responses")
        return self._responses[index]

    def append(self, response: R) -> None:
        self._responses.append(response)

    def items(self) -> list[R]:
        warn(
            "Using items is deprecated and will be removed in v0.5. "
            "Iterate or index the Responses object instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._responses


@dataclass(init=False, repr=False)
class Datum:
    """
    Class to define the important information for the event.
    Args:
        keys: the keys of the event.
        value: the payload of the event.
        event_time: the event time of the event.
        watermark: the watermark of the event.
    >>> # Example usage
    >>> from pynumaflow.sinker import Datum
    >>> from datetime import datetime, timezone
    >>> payload = bytes("test_mock_message", encoding="utf-8")
    >>> t1 = datetime.fromtimestamp(1662998400, timezone.utc)
    >>> t2 = datetime.fromtimestamp(1662998460, timezone.utc)
    >>> msg_id = "test_id"
    >>> output_keys = ["test_key"]
    >>> d = Datum(keys=output_keys, sink_msg_id=msg_id, value=payload, event_time=t1, watermark=t2)
    """

    __slots__ = ("_keys", "_id", "_value", "_event_time", "_watermark")

    _keys: list[str]
    _id: str
    _value: bytes
    _event_time: datetime
    _watermark: datetime

    def __init__(
        self,
        keys: list[str],
        sink_msg_id: str,
        value: bytes,
        event_time: datetime,
        watermark: datetime,
    ):
        self._keys = keys
        self._id = sink_msg_id or ""
        self._value = value or b""
        if not isinstance(event_time, datetime):
            raise TypeError(f"Wrong data type: {type(event_time)} for Datum.event_time")
        self._event_time = event_time
        if not isinstance(watermark, datetime):
            raise TypeError(f"Wrong data type: {type(watermark)} for Datum.watermark")
        self._watermark = watermark

    def __str__(self):
        value_string = self._value.decode("utf-8")
        return (
            f"keys: {self._keys}, "
            f"id: {self._id}, value: {value_string}, "
            f"event_time: {str(self._event_time)}, "
            f"watermark: {str(self._watermark)}"
        )

    def __repr__(self):
        return str(self)

    @property
    def id(self) -> str:
        """Returns the id of the event."""
        return self._id

    @property
    def keys(self) -> list[str]:
        """Returns the keys of the event."""
        return self._keys

    @property
    def value(self) -> bytes:
        """Returns the value of the event."""
        return self._value

    @property
    def event_time(self) -> datetime:
        """Returns the event time of the event."""
        return self._event_time

    @property
    def watermark(self) -> datetime:
        """Returns the watermark of the event."""
        return self._watermark


SinkCallable = Callable[[Iterator[Datum]], Responses]



from pynumaflow.mapstreamer._dtypes import (
    Message,
    Messages,
    Datum,
    DROP,
)
from pynumaflow.mapstreamer.async_server import AsyncMapStreamer

__all__ = [
    "Message",
    "Messages",
    "Datum",
    "DROP",
    "AsyncMapStreamer",
]



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: mapstream.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\x0fmapstream.proto\x12\x0cmapstream.v1\x1a\x1bgoogle/protobuf/empty.proto\x1a\x1fgoogle/protobuf/timestamp.proto"\x8e\x01\n\x10MapStreamRequest\x12\x0c\n\x04keys\x18\x01 \x03(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12.\n\nevent_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12-\n\twatermark\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp"\x80\x01\n\x11MapStreamResponse\x12\x36\n\x06result\x18\x01 \x01(\x0b\x32&.mapstream.v1.MapStreamResponse.Result\x1a\x33\n\x06Result\x12\x0c\n\x04keys\x18\x01 \x03(\t\x12\r\n\x05value\x18\x02 \x01(\x0c\x12\x0c\n\x04tags\x18\x03 \x03(\t"\x1e\n\rReadyResponse\x12\r\n\x05ready\x18\x01 \x01(\x08\x32\x9d\x01\n\tMapStream\x12P\n\x0bMapStreamFn\x12\x1e.mapstream.v1.MapStreamRequest\x1a\x1f.mapstream.v1.MapStreamResponse0\x01\x12>\n\x07IsReady\x12\x16.google.protobuf.Empty\x1a\x1b.mapstream.v1.ReadyResponseb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, "mapstream_pb2", globals())
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    _MAPSTREAMREQUEST._serialized_start = 96
    _MAPSTREAMREQUEST._serialized_end = 238
    _MAPSTREAMRESPONSE._serialized_start = 241
    _MAPSTREAMRESPONSE._serialized_end = 369
    _MAPSTREAMRESPONSE_RESULT._serialized_start = 318
    _MAPSTREAMRESPONSE_RESULT._serialized_end = 369
    _READYRESPONSE._serialized_start = 371
    _READYRESPONSE._serialized_end = 401
    _MAPSTREAM._serialized_start = 404
    _MAPSTREAM._serialized_end = 561
# @@protoc_insertion_point(module_scope)



MAP_SOCK_PATH = "/var/run/numaflow/map.sock"
MAP_STREAM_SOCK_PATH = "/var/run/numaflow/mapstream.sock"
REDUCE_SOCK_PATH = "/var/run/numaflow/reduce.sock"
SOURCE_TRANSFORMER_SOCK_PATH = "/var/run/numaflow/sourcetransform.sock"
SINK_SOCK_PATH = "/var/run/numaflow/sink.sock"
MULTIPROC_MAP_SOCK_PORT = 55551
MULTIPROC_MAP_SOCK_ADDR = "0.0.0.0"
SIDE_INPUT_SOCK_PATH = "/var/run/numaflow/sideinput.sock"

# TODO: need to make sure the DATUM_KEY value is the same as
# https://github.com/numaproj/numaflow-go/blob/main/pkg/function/configs.go#L6
WIN_START_TIME = "x-numaflow-win-start-time"
WIN_END_TIME = "x-numaflow-win-end-time"
MAX_MESSAGE_SIZE = 1024 * 1024 * 64
# TODO: None instead of "EOF" ?
STREAM_EOF = "EOF"
DELIMITER = ":"
DROP = "U+005C__DROP__"
